import pandas as pd
import numpy as np

def stack(layers):
    '''
    Using the Functional-API of Tensorflow to build a sequential
    network (stacked layers) from list of layers.
    '''
    layer_stack = None
    for layer in layers:
        if layer_stack is None:
            layer_stack = layer
        else:
            layer_stack = layer(layer_stack)
    return layer_stack

def create_signal(timestamp,direction,packetsize,relativetime):
    df = pd.DataFrame({'Timestamp':timestamp,'Direction':direction,'PacketSize':packetsize,'RelativeTime':relativetime})
    df['Milliseconds'] = df['RelativeTime'].round().astype(int)
    grouped = df.groupby('Milliseconds').apply(lambda x: x['Direction'].astype(int).sum()).astype(int)
    ms_series = pd.Series(np.zeros(15000, dtype=int), index=np.arange(15000))
    ms_series.update(grouped)
    return np.stack(ms_series.values.tolist())


from .stnn import STNN
from .clumps_flow import ClumpsFlow
from .packets_size_interarrival_time import PacketsSizeAndIAT
import pandas as pd
import numpy as np

class STNNExtended():
    ''' STNNExtended |
    Extracts 140 features from `n_packets` packets. The extended version
    maintain only flow direction statistical features src2dst, dst2src the same and
    for the bidectional take only the packets,bytes,packets per second and bytes per second.
    Which are total of 32 features. In addition added 108 new features from clumps flow 84 feautres
    and from packet size and inter-arrival-time bidrectional flow 24 features a total of 140 features.
    
    Feature Outputs:
        - udps.stnn_image_enh: output shape of (140).
    '''
    size_iat_cols = [
            'udps.packets_size_min',
            'udps.packets_size_max',
            'udps.packets_size_stddev',
            'udps.packets_size_first_quartile',
            'udps.packets_size_second_quartile',
            'udps.packets_size_third_quartile',
            'udps.packets_size_mean',
            'udps.packets_size_median_absoulte_deviation',
            'udps.packets_size_variance',
            'udps.packets_size_skewness',
            'udps.packets_size_kurtosis',
            'udps.packets_size_sum',
            'udps.packets_interarrival_time_min',
            'udps.packets_interarrival_time_max',
            'udps.packets_interarrival_time_stddev',
            'udps.packets_interarrival_time_first_quartile',
            'udps.packets_interarrival_time_second_quartile',
            'udps.packets_interarrival_time_third_quartile',
            'udps.packets_interarrival_time_mean',
            'udps.packets_interarrival_time_median_absoulte_deviation',
            'udps.packets_interarrival_time_variance',
            'udps.packets_interarrival_time_skewness',
            'udps.packets_interarrival_time_kurtosis',
            'udps.packets_interarrival_time_sum'
        ] # 24

    clump_cols = ['udps.src2dst_max_clumps_len',
        'udps.src2dst_min_clumps_len',
        'udps.src2dst_mean_clumps_len',
        'udps.src2dst_stddev_clumps_len',
        'udps.src2dst_skewness_clumps_len',
        'udps.src2dst_variance_clumps_len',
        'udps.src2dst_kurtosis_clumps_len',
        'udps.src2dst_max_clumps_size',
        'udps.src2dst_min_clumps_size',
        'udps.src2dst_mean_clumps_size',
        'udps.src2dst_stddev_clumps_size',
        'udps.src2dst_skewness_clumps_size',
        'udps.src2dst_variance_clumps_size',
        'udps.src2dst_kurtosis_clumps_size',
        'udps.src2dst_max_clumps_bytes_per_packet',
        'udps.src2dst_min_clumps_bytes_per_packet',
        'udps.src2dst_mean_clumps_bytes_per_packet',
        'udps.src2dst_stddev_clumps_bytes_per_packet',
        'udps.src2dst_skewness_clumps_bytes_per_packet',
        'udps.src2dst_variance_clumps_bytes_per_packet',
        'udps.src2dst_kurtosis_clumps_bytes_per_packet',
        'udps.src2dst_max_clumps_interarrival_time',
        'udps.src2dst_min_clumps_interarrival_time',
        'udps.src2dst_mean_clumps_interarrival_time',
        'udps.src2dst_stddev_clumps_interarrival_time',
        'udps.src2dst_skewness_clumps_interarrival_time',
        'udps.src2dst_variance_clumps_interarrival_time',
        'udps.src2dst_kurtosis_clumps_interarrival_time',
        'udps.dst2src_max_clumps_len',
        'udps.dst2src_min_clumps_len',
        'udps.dst2src_mean_clumps_len',
        'udps.dst2src_stddev_clumps_len',
        'udps.dst2src_skewness_clumps_len',
        'udps.dst2src_variance_clumps_len',
        'udps.dst2src_kurtosis_clumps_len',
        'udps.dst2src_max_clumps_size',
        'udps.dst2src_min_clumps_size',
        'udps.dst2src_mean_clumps_size',
        'udps.dst2src_stddev_clumps_size',
        'udps.dst2src_skewness_clumps_size',
        'udps.dst2src_variance_clumps_size',
        'udps.dst2src_kurtosis_clumps_size',
        'udps.dst2src_max_clumps_bytes_per_packet',
        'udps.dst2src_min_clumps_bytes_per_packet',
        'udps.dst2src_mean_clumps_bytes_per_packet',
        'udps.dst2src_stddev_clumps_bytes_per_packet',
        'udps.dst2src_skewness_clumps_bytes_per_packet',
        'udps.dst2src_variance_clumps_bytes_per_packet',
        'udps.dst2src_kurtosis_clumps_bytes_per_packet',
        'udps.dst2src_max_clumps_interarrival_time',
        'udps.dst2src_min_clumps_interarrival_time',
        'udps.dst2src_mean_clumps_interarrival_time',
        'udps.dst2src_stddev_clumps_interarrival_time',
        'udps.dst2src_skewness_clumps_interarrival_time',
        'udps.dst2src_variance_clumps_interarrival_time',
        'udps.dst2src_kurtosis_clumps_interarrival_time',
        'udps.max_clumps_len',
        'udps.min_clumps_len',
        'udps.mean_clumps_len',
        'udps.stddev_clumps_len',
        'udps.skewness_clumps_len',
        'udps.variance_clumps_len',
        'udps.kurtosis_clumps_len',
        'udps.max_clumps_size',
        'udps.min_clumps_size',
        'udps.mean_clumps_size',
        'udps.stddev_clumps_size',
        'udps.skewness_clumps_size',
        'udps.variance_clumps_size',
        'udps.kurtosis_clumps_size',
        'udps.max_clumps_bytes_per_packet',
        'udps.min_clumps_bytes_per_packet',
        'udps.mean_clumps_bytes_per_packet',
        'udps.stddev_clumps_bytes_per_packet',
        'udps.skewness_clumps_bytes_per_packet',
        'udps.variance_clumps_bytes_per_packet',
        'udps.kurtosis_clumps_bytes_per_packet',
        'udps.max_clumps_interarrival_time',
        'udps.min_clumps_interarrival_time',
        'udps.mean_clumps_interarrival_time',
        'udps.stddev_clumps_interarrival_time',
        'udps.skewness_clumps_interarrival_time',
        'udps.variance_clumps_interarrival_time',
        'udps.kurtosis_clumps_interarrival_time'] # 84



    def plugins(n_packets=32):
        return ClumpsFlow(n_packets),STNN(n_packets),PacketsSizeAndIAT(n_packets)

    @staticmethod
    def stnn_feat_enh(row):

        feat = row[STNNExtended.size_iat_cols + STNNExtended.clump_cols]
        sub_stnn =  np.concatenate([np.asarray(row['udps.stnn_image'][0][10:]),np.asarray(row['udps.stnn_image'][1:3]).flatten()])
        return np.concatenate([sub_stnn, feat]).astype('float32')

            
    @staticmethod
    def preprocess(dataframe: pd.DataFrame):
        ''' 
        Preprocessing method for the STNN extended features.
        Converting 'udps.stnn_image_enh' column from str to 1D-list.
        import
        '''
        # validate
        STNN.preprocess(dataframe)

        dataframe['udps.stnn_image_enh'] = dataframe.apply(STNNExtended.stnn_feat_enh,axis=1)
    


from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input,Layer,Concatenate,Dense,BatchNormalization,Dropout
from tensorflow.keras import activations
import tensorflow as tf

class GraphDAppModality:
    '''
    GraphDApp deep learning nerual network uses for Graph input.
    Classifciation type: Graphs
    Input: vector of two values an adjacent matrix and a node feature vector.

    Extraction plugin: SimpleTIG

    Paper: "Accurate Decentralized Application Identification via Encrypted Traffic Analysis Using Graph Neural Networks."
    Authors: Meng Shen, Jinpeng Zhang, Liehuang Zhu, Ke Xu, and Xiaojiang Du.
    
    '''
    def __init__(self,n_packets=32) -> None:
        input_layer_graphdapp_adj_matrix_modality = Input(shape=(n_packets,n_packets),name='input_graphdapp_adj_matrix')
        input_layer_graphdapp_features_modality = Input(shape=(n_packets,1),name='input_graphdapp_features')
        inputs_layer = [input_layer_graphdapp_adj_matrix_modality, input_layer_graphdapp_features_modality]
        mlp1 = MLPLayer(64)
        mlp2 = MLPLayer(64)
        mlp3 = MLPLayer(64)
        readout = Readout()
        concat = Concatenate()
        x1 = mlp1(inputs_layer)
        x2 = mlp2(x1)
        x3 = mlp3(x2)
        x4 = readout([x1, x2, x3])
        x4 = concat(x4)
        self.model = Model(
            name='GraphDApp modality',
            inputs=inputs_layer,
            outputs= x4
        )

class Readout(Layer):
    def __init__(self):
        super(Readout, self).__init__()
    
    def call(self, inputs):
        FsBatch = []
        for i in range(len(inputs)):
            Fbatch = inputs[i][1]
            FsBatch.append(tf.reduce_sum(Fbatch, axis=-2))
        return FsBatch


class MLPLayer(Layer):
    def __init__(
            self, 
            output_size=5, 
            activation='relu', 
            use_bias=True, 
            neighbour_agg_method='sum', 
            dropout_rate=0.025):
        super(MLPLayer, self).__init__()
        self.output_size = output_size
        self.dense = Dense(self.output_size, use_bias=use_bias)
        self.activation = activations.get(activation)
        self.batch_norm = BatchNormalization()
        self.dropout = Dropout(dropout_rate)
    
    def build(self, input_shape):
        self.eps = self.add_weight(
            name='epsilon',
            shape=(1,),
            initializer="random_normal",
            dtype='float32',
            trainable=True,
        )
    
    def call(self, inputs, training=False):
        '''
        [A, F]
        '''
        A = inputs[0]
        F = inputs[1]
        outputs = tf.multiply(1.0+self.eps, F) + tf.matmul(A,F)
        outputs = self.dense(outputs)
        outputs = self.activation(outputs)
        outputs = self.batch_norm(outputs, training=training)
        outputs = self.dropout(outputs)
        return [A, outputs]



        
class HybridPreprocessing:
    @staticmethod
    def preprocess_features(dataframe):
        from EIMTC.plugins.n_bytes import NBytes
        from EIMTC.plugins.protocol_header_fields_extended import ProtocolHeaderFields_Extended
        from EIMTC.plugins.stnn_extended import STNN_Extended
        from EIMTC.plugins.simple_tig import SimpleTIG
        NBytes.preprocess(dataframe)
        ProtocolHeaderFields_Extended.preprocess(dataframe)
        STNN_Extended.preprocess(dataframe)
        SimpleTIG.preprocess(dataframe)


from EIMTC.metrics import accuracy_score
import numpy as np

def multi_label_accuracy_score(y_true, y_pred, **kwargs):
    scores = []
    for true, pred in zip(y_true, y_pred):
        scores.append(accuracy_score(true, pred))
    
    return tuple(scores)


def multi_label_tuple_accuracy_score(y_true, y_pred):
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    score = np.sum(np.all(y_true.T == y_pred.T, axis=1))

    match_count = len(score[score == True])
    sample_count = y_true.shape[1]
    accuracy = match_count/sample_count
    
    return accuracy


def multi_label_tuple_matrix(y_true, y_pred):
    results_per_class = []
    for combination in np.stack(np.meshgrid(*[np.unique(lbls) for lbls in y_true]), -1).reshape(-1, len(y_true)):
        result_dict = multi_label_tuple_for_class(y_true, y_pred, combination)
        result_dict['class'] = combination
        results_per_class.append(result_dict)
        
    return results_per_class

def multi_label_tuple_for_class(y_true, y_pred, classes):
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    
    # tp
    samples_indices_for_class = np.where((y_true.T == classes).all(axis=1))[0]
    predictions_for_indices = y_pred.T[samples_indices_for_class]
    tp = np.sum(np.all(predictions_for_indices == classes, axis=1))
    # fn
    fn = np.sum(np.any(predictions_for_indices != classes, axis=1))
    # fp
    samples_indices_for_not_class = np.where((y_true.T != classes).any(axis=1))[0]
    predictions_for_indices = y_pred.T[samples_indices_for_not_class]
    fp = np.sum(np.all(predictions_for_indices == classes, axis=1))
    # tn
    tn = np.sum(np.any(predictions_for_indices != classes, axis=1))
    
    return {
        'tp': tp,
        'fp': fp,
        'tn': tn,
        'fn': fn,
    }
    
def multi_label_tuple_recall_score(y_true, y_pred, mode='class'):
    if mode == 'class':
        return multi_label_tuple_recall_score_per_tuple(y_true, y_pred)
    elif mode == 'macro':
        return multi_label_tuple_recall_score_macro(y_true, y_pred)

def multi_label_tuple_recall_score_per_tuple(y_true, y_pred):
    mat = multi_label_tuple_matrix(y_true, y_pred)
    recalls = []
    for cls in mat:
        if (cls['tp'] + cls['fn']) == 0:
            recall_for_class = 0.0
        else:
            recall_for_class = cls['tp'] / (cls['tp'] + cls['fn'])
        recalls.append(recall_for_class)
    
    return recalls
        
def multi_label_tuple_recall_score_macro(y_true, y_pred):
    recalls = multi_label_tuple_recall_score_per_tuple(y_true, y_pred)
    recall_score = np.sum(recalls)/len(recalls)
    
    return recall_score


## PRECISION ##
def multi_label_tuple_precision_score(y_true, y_pred, mode='class'):
    if mode == 'class':
        return multi_label_tuple_precision_score_per_tuple(y_true, y_pred)
    elif mode == 'macro':
        return multi_label_tuple_precision_score_macro(y_true, y_pred)

def multi_label_tuple_precision_score_per_tuple(y_true, y_pred, mode='class'):
    mat = multi_label_tuple_matrix(y_true, y_pred)
    precisions = []
    for cls in mat:
        if (cls['tp'] + cls['fp']) == 0:
            recall_for_class = 0.0
        else:
            recall_for_class = cls['tp'] / (cls['tp'] + cls['fp'])
        precisions.append(recall_for_class)    
        
    return precisions

def multi_label_tuple_precision_score_macro(y_true, y_pred):
    precisions = multi_label_tuple_precision_score_per_tuple(y_true, y_pred)
    precision_score = np.sum(precisions)/len(precisions)
    
    return precision_score


## F1 ##
def multi_label_tuple_f1_score(y_true, y_pred, mode='class'):
    if mode == 'class':
        return multi_label_tuple_f1_score_per_tuple(y_true, y_pred)
    elif mode == 'macro':
        return multi_label_tuple_f1_score_macro(y_true, y_pred)

def multi_label_tuple_f1_score_per_tuple(y_true, y_pred):
    mat_recall = np.asarray(multi_label_tuple_recall_score(y_true, y_pred))
    mat_precision = np.asarray(multi_label_tuple_precision_score(y_true, y_pred))
    f1_per_class = np.nan_to_num((2*mat_recall*mat_precision)/(mat_recall+mat_precision))
    
    return f1_per_class

def multi_label_tuple_f1_score_macro(y_true, y_pred):
    f1s = multi_label_tuple_f1_score_per_tuple(y_true, y_pred)
    f1_score = np.sum(f1s)/len(f1s)
    
    return f1_score


import pywt
import numpy as np
from ._modal_wrapper import ModalWrapper
from sklearn.ensemble import RandomForestClassifier

class WaveletModality:
    '''
    Wavelet is a deep learning nerual network which uses waves signals with wavelet of the packets flow as input.
    Classifciation type: Time Series
    Input: runs on 15000 miliseconds using time and direction as it features

    Contributor: Natan Dilbary
    '''
    def __init__(self,input_shape) -> None:
        self.model = ModalWrapper(RandomForestClassifier(),name='wavelet_modality',input_shape=input_shape)
    
    def fit_transform(self,x):
        return self.transform_signals2Wavelet(x)
    
    def extract_wavelet_features(self,data, features_to_extract=None, wavelet='coif6', level=4):
        # Compute the DWT of your signal
        coeffs = pywt.wavedec(data, wavelet, level=level)
        epsilon = 1e-10  # small constant to avoid division by zero

        feature_map = {
            'mean': lambda x: np.mean(x),
            'std': lambda x: np.std(x),
            'median': lambda x: np.median(x),
            'max': lambda x: np.max(x),
            'min': lambda x: np.min(x),
            'range': lambda x: np.max(x) - np.min(x),
            'energy': lambda x: np.sum(np.square(x)),
            'crest_factor': lambda x: np.max(np.abs(x)) / (np.sqrt(np.mean(np.square(x))) + epsilon),
            'shape_factor': lambda x: np.sqrt(np.mean(np.square(x))) / (np.mean(np.abs(x)) + epsilon),
        }

    # If no features are specified, use all features
        if not features_to_extract:
            features_to_extract = list(feature_map.keys())

        # Compute the requested features
        computed_features = []
        for coeff in coeffs:
            computed_features.extend([feature_map[feat](coeff) for feat in features_to_extract if feat in feature_map])

        # Assemble the feature vector
        feature_vec = np.array(computed_features)
        return feature_vec
    
    def transform_signals2Wavelet(self,signals):
        wavelet_list = []
        for signal in signals:
            features = np.array([self.extract_wavelet_features(signal)])
            wavelet_list.append(features)
        return np.concatenate(wavelet_list)
    



from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input,GRU,Bidirectional,ReLU,Flatten
from tensorflow.keras.constraints import max_norm 
from ._utils import stack

class LopezTDLModality:
    '''
    Lopez-Protocol-header-Fields TDL deep learning nerual network uses for Packet Header Fields vector input.
    Classifciation type: Packet Header Fields
    Input: two-dimensional vector of number of packets and 3 features from each packet.

    Extraction plugin: ProtocolHeaderFieldsExtended
    
    Inspired by the paper: "Network Traffic Classifier With Convolutional and Recurrent Neural Networks for Internet of Things"
    '''
    def __init__(self,packet_count=32) -> None:
        input_layer_protocol_fields_modality = Input(shape=(packet_count,3), name='input_protocol_fields_TDL')
        self.model = Model(
            name='Lopez protocol header fields TDL modality',
            inputs=input_layer_protocol_fields_modality,
            outputs=stack([
                input_layer_protocol_fields_modality,
                Bidirectional(GRU(64, return_sequences=True, kernel_constraint=max_norm(3))),
                ReLU(),
                Flatten(),
            ])
        )


from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input,Conv1D,ReLU,MaxPooling1D,Flatten
from ._utils import stack

class WangPayloadModality:
    '''
    WangPayload deep learning nerual network uses for 1D vector input.
    Classifciation type: Payload
    Input: vector 1D of 784 bytes as features.

    Extraction plugin: NBytes

    Paper: "End-to-end Encrypted Traffic Classification with One-dimensional Convolution Neural Networks."
    Authors: Wei Wang, Ming Zhu, Jinlin Wang, Xuewen Zeng, and Zhongzhen Yang.
    '''
    def __init__(self,payload_size=784) -> None:
        input_layer_payload_modality = Input(shape=(payload_size,1), name='input_payload')
        self.model = Model(
            name='Wang payload modality - nbytes',
            inputs=input_layer_payload_modality,
            outputs=stack([
                input_layer_payload_modality,
                Conv1D(16, 25, name='Conv1D_payload_1'),
                ReLU(),
                MaxPooling1D(3, name='MaxPooling1D_payload_1'),
                Conv1D(32, 35, name='Conv1D_payload_2'),
                ReLU(),
                MaxPooling1D(3, name='MaxPooling1D_payload_2'),
                Flatten(), 
            ])
        )


from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input,Lambda,LeakyReLU,Bidirectional,Flatten,Conv2D,MaxPool2D,LSTM,Dense
from tensorflow import expand_dims
from ._utils import stack


class STNNModality:
    '''
    STNN Exteneded deep learning nerual network uses for 2D matrix input.
    Classifciation type: packets and flow statistic
    Input: two-dimensional matrix of 70 features in the shape of 5 x 14.

    Extraction plugin: STNN
    
    Inspired by the paper: "Network Traffic Classifier With Convolutional and Recurrent Neural Networks for Internet of Things"
    '''
    def __init__(self) -> None:
        input_layer_stnn_modality = Input(shape=(5,14), name='input_stnn')
        self.model = Model(
            name='STNN-inspired image modality',
            inputs=input_layer_stnn_modality,
            outputs=stack([
                input_layer_stnn_modality,
                Bidirectional(LSTM(65,return_sequences=True)),
                Lambda(lambda x: expand_dims(x, axis=3)),
                Conv2D(32,3,padding='same'),
                LeakyReLU(),
                Conv2D(32,3,padding='same'),
                LeakyReLU(),
                MaxPool2D(2),
                Conv2D(64,3,padding='same'),
                LeakyReLU(),
                Conv2D(128,3,padding='same'),
                LeakyReLU(),
                MaxPool2D(2),
                Flatten(),
                Dense(512),
            ])
        )



import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # or any {'0', '1', '2'}
import tensorflow as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.WARN)  # or any {DEBUG, INFO, WARN, ERROR, FATAL}
from tensorflow.keras.layers import *
from tensorflow.keras.models import *
from tensorflow.python.trackable.data_structures import ListWrapper




class CustomDistiller(Model):
    def __init__(self, modalities=[], adapter_size=128, n_classes=[], merging_method='feat_concat',adapters_setup=[]) -> None:
        super(CustomDistiller, self).__init__()
        self.n_classes = n_classes
        self.modalities = modalities
        self.adapter_size = adapter_size
        self.merging_method = merging_method
        self.adapters_setup = adapters_setup
        self._validate()
        
        
        if self.merging_method == 'feat_concat':
            shared_representation = feat_concat_method(get_sub_adapters(modalities,adapter_size,merging_method,adapters_setup),adapter_size)

        elif self.merging_method == 'deep_concat':
            shared_representation = deep_concat_method(get_sub_adapters(modalities,adapter_size,merging_method,adapters_setup),adapter_size)
            
        elif self.merging_method == 'feat_merge':
            shared_representation = feat_merge_method(get_sub_adapters(modalities,adapter_size,merging_method,adapters_setup),adapter_size)
        
        elif self.merging_method == 'feat_cnn':
            shared_representation = feat_cnn_method(get_sub_adapters(modalities,adapter_size,merging_method,adapters_setup),adapter_size)
        
        outputs = []
        for n_class in n_classes:
            outputs.append(stack([shared_representation] + get_ts_layers(classes_count=n_class, adapter_size=adapter_size)))
        self.model = Model(
            name='CustomDistiller',
            inputs=[modal.input for modal in modalities],
            outputs= outputs
        )
        
        self.pretraining_models = [self.get_model_for_pretraining(modal) for modal in modalities]


    def call(self, inputs, training=None):
        # See: https://www.tensorflow.org/guide/keras/custom_layers_and_models#the_model_class
        return self.model(inputs, training)


    def compile(self, **kwargs):
        self.compile_settings = kwargs
        for model, compile_kwargs in zip(self.pretraining_models + [self.model], zip_dict(kwargs)):
            model.compile(**compile_kwargs)

    def _compile(self, **kwargs): # compile for freezing and unfreezing
        for model, compile_kwargs in zip(self.pretraining_models + [self.model], zip_dict(kwargs)):
            model.compile(**compile_kwargs)

    def fit(self, x, y, **kwargs):
        kwargs_per_fit = list(zip_dict(kwargs))
    
        for features, model, fit_kwargs in zip(x, self.pretraining_models, kwargs_per_fit):
            print('##################### {} ##########################'.format(model.name.upper()))
            model.fit(features, y, **fit_kwargs)

        # FINE-TUNE
        print('##################### FINE-TUNING ##########################')
        fit_kwargs = kwargs_per_fit[len(self.modalities)]
        self.freeze_for_finetuning()
        self.model.fit(x,y, **fit_kwargs)
        self.unfreeze_for_finetuning()

    def fit_display(self, x, y,queue, **kwargs):
        kwargs_per_fit = list(zip_dict(kwargs))
    
        for features, model, fit_kwargs in zip(x, self.pretraining_models, kwargs_per_fit):
            queue.put('##################### {} ##########################'.format(model.name.upper()))
            model.fit(features, y, **fit_kwargs)

        # FINE-TUNE
        queue.put('##################### FINE-TUNING ##########################')
        fit_kwargs = kwargs_per_fit[len(self.modalities)]
        self.freeze_for_finetuning()
        self.model.fit(x,y, **fit_kwargs)
        self.unfreeze_for_finetuning()

    def modals_select_fit(self,x,y,num_modals,to_train=True,**kwargs): # new selected modal allows to choose which modals to pre-train with some modals already pre-train
        kwargs_per_fit = list(zip_dict(kwargs))

        if to_train:
            untrained_modals = [self.pretraining_models[i] for i in num_modals]
        else:
            num_m = []
            for i in range(len(self.pretraining_models)):
                if i not in num_modals:
                    num_m.append(i)
            untrained_modals = [self.pretraining_models[i] for i in num_m]
        
        for features, model, fit_kwargs in zip(x, untrained_modals, kwargs_per_fit):
            print('##################### {} ##########################'.format(model.name.upper()))
            model.fit(features, y, **fit_kwargs)

        # FINE-TUNE
        print('##################### FINE-TUNING ##########################')
        fit_kwargs = kwargs_per_fit[len(self.modalities)]
        self.freeze_for_finetuning()
        self.model.fit(x,y, **fit_kwargs)
        self.unfreeze_for_finetuning()

    def _validate(self):
        # modalities and n_classess array must not be empty .
        assert len(self.modalities) != 0
        assert len(self.n_classes) != 0
        
        # n_classes values must represent the number of classes in each task.
        # and therefore must be positive integers.
        for nclass in self.n_classes:
            assert nclass >= 1
        
        # adapter_size must be positive.
        assert self.adapter_size >= 1


    def get_model_for_pretraining(self, model):
        model_w_adapter = wrap_adapter(model, self.adapter_size)
        outputs = []
        for n_class in self.n_classes:
            outputs.append(stack([model_w_adapter, Dense(n_class, activation='softmax')]))
        return Model(
            name='_'.join(['pretraining_model', model.name.replace(' ','_')]),
            inputs=model.input,
            outputs=outputs
        )


    def freeze_for_finetuning(self):
        for modal in self.modalities:
            for layer in modal.layers:
                layer.trainable = False
        self._compile(**self.compile_settings)


    def unfreeze_for_finetuning(self):
        for modal in self.modalities:
            for layer in modal.layers:
                layer.trainable = True
        self._compile(**self.compile_settings)
        
    



###################
# Model utilities #
###################


def stack(layers):
    '''
    Using the Functional-API of Tensorflow to build a sequential
    network (stacked layers) from list of layers.
    '''
    layer_stack = None
    for layer in layers:
        if layer_stack is None:
            layer_stack = layer
        else:
            layer_stack = layer(layer_stack)
    return layer_stack


def get_adapter_layers(adapter_size):
    return [
        Dropout(0.2),
        Dense(adapter_size),
        ReLU()
    ]


def wrap_adapter(model, adapter_size):
    return stack([model.output, *get_adapter_layers(adapter_size)])


def wrap_adapter_multi(models, adapter_size):
    return [wrap_adapter(model, adapter_size) for model in models]


def get_sr_layers(adapter_size):
    # SR = Shared Representation
    return [
        Dropout(0.2),
        Dense(adapter_size),
        ReLU(),
        Dropout(0.2),
    ]


def get_ts_layers(classes_count, adapter_size):
    # TS = Task Specific
    return [
        Dense(adapter_size),
        ReLU(),
        Dropout(0.2),
        Dense(classes_count),
        Softmax()
    ]
    
def feat_concat_method(modalities,adapter_size):
    return stack(
                [
                    Concatenate()(
                        wrap_adapter_multi(modalities, adapter_size) 
                    )
                ]
                + get_sr_layers(adapter_size)
            )

def deep_concat_method(modalities,adapter_size):
    return stack(
                [
                    Concatenate()(
                        [model.output for model in modalities]
                    )
                ]
                + get_sr_layers(adapter_size)
            )

def feat_merge_method(modalities,adapter_size):
    return stack(
                [
                    Add()(
                        wrap_adapter_multi(modalities, adapter_size) 
                    )
                ]
                + get_sr_layers(adapter_size)
            )
    
def feat_cnn_method(modalities,adapter_size):
    return stack(
        [
                Concatenate()(
                        wrap_adapter_multi(modalities, adapter_size) 
                ),
               Reshape((len(modalities),-1,1)),
               Conv2D(16,3,strides=(1,1),padding="same",activation='relu'),
               MaxPooling2D(3, strides=(3,3),padding="same"),
               Conv2D(32,3,strides=(1,1),padding="same",activation='relu'),
               MaxPooling2D(3, strides=(3,3),padding="same"),
               Conv2D(64,3,strides=(1,1),padding="same",activation='relu'),
               MaxPooling2D(3, strides=(3,3),padding="same"),
               Flatten(),
               Dense(64,activation='relu'),
               Dropout(0.2),
        ]
        + get_sr_layers(adapter_size)
    )
    
def get_sub_adapter(models,adapter_size,merging_method):
    if merging_method == 'deep_concat':
        return Model(
            inputs=[modal.input for modal in models],
            outputs= deep_concat_method(models,adapter_size)
        )
    if merging_method == 'feat_merge':
        return Model(
            inputs=[modal.input for modal in models],
            outputs= feat_merge_method(models,adapter_size)
        )
    if merging_method == 'feat_cnn':
        return Model(
            inputs=[modal.input for modal in models],
            outputs= feat_cnn_method(models,adapter_size)
        )
    return Model(
            inputs=[modal.input for modal in models],
            outputs= feat_concat_method(models,adapter_size)
        )

def get_sub_adapters(models,adapter_size,merging_method,adapters_setup):
    if not adapters_setup:
        return models
    adapters = []
    for adapter in adapters_setup:
        if type(adapter) is ListWrapper:
            adapters.append(get_sub_adapter([models[i] for i in adapter],adapter_size,merging_method))
        else:
            adapters.append(models[adapter])
    return adapters

def zip_dict(d):
    for vals in zip(*(d.values())):
        yield dict(zip(d.keys(), vals))
        


from nfstream import NFPlugin

class TDL(NFPlugin):
    '''TDL(Time,packet-Direction,packet Length) | A 2D matrix of extract Time,packet-direction and packet length from each packet.
    
    Output Features:
        Select all packets from each flow and extract Time,packet-direction and packet length from each packet.

        ```
    
    '''
    def __init__(self, flow_time=None, n_packet=None):
        ''' '''
        self.flow_time = flow_time
        self.n_packet = n_packet

        self.flow_time_flag = False
        self.n_packet_flag = False
        self.time_packet_flag = False
        if self.flow_time is not None and self.n_packet is not None: # flow has to be with at least flow_time seconds and at least with n_packets
            self.time_packet_flag = True
        elif self.flow_time is not None:
            self.flow_time_flag = True
        elif self.n_packet is not None:
            self.n_packet_flag = True

    def on_init(self, packet, flow):
        ''' '''
        flow.udps.port_TDL= list()
        flow.udps.ip_TDL= list()
        src_port = packet.src_port <= 1024
        dst_port = packet.dst_port <= 1024
        # session direction by port
        if src_port and not dst_port: # 1,0
            flow.udps.port_direction = 1
        elif not src_port and dst_port:# 0,1
            flow.udps.port_direction = 0
        else:# 0,0 and 1,1
            if packet.src_port < packet.dst_port:# check which port is smaller
                flow.udps.port_direction = 1
            else:
                flow.udps.port_direction = 0
        if flow.ip_version == 4: # dst ip is ipv4
            #src_ip = packet.src_ip.split('.')
            dst_ip = packet.dst_ip.split('.')
            if (int(dst_ip[0]) == 10 or (int(dst_ip[0])== 172 and 16 <= int(dst_ip[1]) and int(dst_ip[1]) <= 31) or (int(dst_ip[0]) == 192 and int(dst_ip[1]) == 168)):
                flow.udps.ip_direction = 1
            else:
                flow.udps.ip_direction = 0
        if flow.ip_version == 6:# dst ip is ipv6
            #src_ip = packet.src_ip[:packet.src_ip.find(':')]
            dst_ip = packet.dst_ip[:packet.dst_ip.find(':')]
            if dst_ip == 'fe80' or dst_ip == 'fc00' or dst_ip == 'fec0':
                flow.udps.ip_direction = 1
            else:
                flow.udps.ip_direction = 0
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        ''' '''
        if self.time_packet_flag and (packet.time - flow.bidirectional_first_seen_ms) > self.flow_time and flow.bidirectional_packets > self.n_packet:
            return
        
        if self.flow_time_flag and (packet.time - flow.bidirectional_first_seen_ms) > self.flow_time:
            return
        
        if self.n_packet_flag and flow.bidirectional_packets > self.n_packet:
            return
        
        # s_packet = Ether()
        if packet.protocol in [6, 17]: # TCP or UDP only.
            vf_port = list()
            vf_ip = list()
            vf_port.append(packet.time/1000) # timestamp (seconds)
            vf_ip.append(packet.time/1000) # timestamp (seconds)
            #vf.append(packet.delta_time) # inter arrival time
            #vf.append(packet.time - flow.bidirectional_first_seen_ms) # relative time
            
            if flow.udps.port_direction == 0:    
                vf_port.append(abs(packet.direction-1))# packet direction
            else:
                vf_port.append(packet.direction)# packet direction

            if flow.udps.ip_direction == 0:    
                vf_ip.append(abs(packet.direction-1))# packet direction
            else:
                vf_ip.append(packet.direction)# packet direction

            vf_port.append(packet.ip_size)# packet size
            vf_ip.append(packet.ip_size)# packet size

            # vf_port.append(packet.time - flow.bidirectional_first_seen_ms) # relative time
            # vf_ip.append(packet.time - flow.bidirectional_first_seen_ms) # relative time

            flow.udps.port_TDL.append(vf_port)# append packet feature vector to list
            flow.udps.ip_TDL.append(vf_ip)# append packet feature vector to list
                
    def on_expire(self,flow):
        del flow.udps.port_direction
        del flow.udps.ip_direction
    
        
    



import tensorflow as tf
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Layer, BatchNormalization, ReLU, Dense,  Dropout

class ModalWrapper(tf.keras.Model):
    '''
    ModelWrapper is a deep learning nerual network given a inner model to train and learn on.
    Classifciation type: Any
    Input: changes by the inner model

    Contributor: Adi Lichy
    '''
    def __init__(self,inner_model,name,input_shape=None,deep_model=None,units=128,dropout_rate=0.2):
        super(ModalWrapper, self).__init__(name=name)
        self.inner_model = inner_model
        self.input_shape_wrapper = input_shape
        self.deep_model = deep_model
        self._validate()
        # Layers
        if self.prebuilt_layers:
            self.input_layer = Input(shape=self.input_shape_wrapper,name=f'input_{name.replace("_modality","")}')
            self.dense1 = Dense(units)
            self.activation1 = ReLU()
            self.batch_norm1 = BatchNormalization()
            self.dropout1 = Dropout(dropout_rate)
            self.dense2 = Dense(units)
            self.activation2 = ReLU()
            self.batch_norm2 = BatchNormalization()
            self.dropout2 = Dropout(dropout_rate)
            self(inputs=self.input_layer)
        else:
            self(inputs=deep_model.inputs)


    # validating modal parameters
    def _validate(self):

        assert hasattr(self.inner_model,'fit') # must have fit to pass the data

        if hasattr(self.inner_model,'predict'):
            self.is_classifer = True
        else:
            self.is_classifer = False

        assert not (self.input_shape_wrapper is None and self.deep_model is None) # check for nerual network build

        if self.input_shape_wrapper is not None:
            self.prebuilt_layers = True
        else:
            self.prebuilt_layers = False

    def fit_wrapper_model(self,x,y=None):
        if y is not None:
            self.inner_model.fit(x,y)
        else:
            self.inner_model.fit(x)
    
    def predict_wrapper_model(self,x):
        return self.inner_model.predict(x)

    def call(self,inputs):
        if self.prebuilt_layers:
            dense1_out = self.dense1(inputs)
            acti1_out = self.activation1(dense1_out)
            norm1_out = self.batch_norm1(acti1_out)
            dropout1_out = self.dropout1(norm1_out)
            dense2_out = self.dense2(dropout1_out)
            acti2_out = self.activation2(dense2_out)
            norm2_out = self.batch_norm2(acti2_out)
            dropout2_out = self.dropout2(norm2_out)
            return  dropout2_out
        else:
            return self.deep_model(inputs)
    
    def fit(self,x=None,y=None,**kwargs):
        if self.is_classifer:
            print('Inner model fit...')
            self.fit_wrapper_model(x,y)
            inner_model_x = np.stack(self.predict_wrapper_model(x))
        else:
            print('Inner model fit...')
            inner_model_x = self.fit_wrapper_model(x)
        print('Deep model fit...')
        return super(ModalWrapper,self).fit(inner_model_x,np.stack(y),**kwargs)
    
    def predict(self,x,**kwargs):
        if self.is_classifer:
            inner_model_x = np.stack(self.predict_wrapper_model(x))
        else:
            inner_model_x = self.fit_wrapper_model(x)
        return super(ModalWrapper,self).predict(inner_model_x,**kwargs)


from sklearn.model_selection import (
    train_test_split, KFold, StratifiedKFold, GroupKFold, ShuffleSplit
)


__all__ = [
    'train_test_split',
    'KFold', 
    'StratifiedKFold', 
    'GroupKFold', 
    'ShuffleSplit',
]


from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input,LeakyReLU,Bidirectional,Flatten,Conv1D,MaxPool1D,LSTM,Dense
from ._utils import stack

class STNNExtendedModality:
    '''
    STNN Exteneded deep learning nerual network uses for 1D vector input.
    Classifciation type: packets and flow statistic
    Input: one-dimensional vector of 140 features.

    Extraction plugin: STNNExtended
    
    Inspired by the paper: "Network Traffic Classifier With Convolutional and Recurrent Neural Networks for Internet of Things"
    '''
    def __init__(self) -> None:
        input_layer_stnn_modality = Input(shape=(140,1), name='input_stnn')
        self.model = Model(
            name='STNN-inspired image extended modality',
            inputs=input_layer_stnn_modality,
            outputs=stack([
                input_layer_stnn_modality,
                Bidirectional(LSTM(64,return_sequences=True)), #128
                Conv1D(32,3,padding='same'),
                LeakyReLU(),
                #Conv1D(32,3,padding='same'),
                #LeakyReLU(),
                #MaxPool1D(2),
                Conv1D(64,3,padding='same'),
                LeakyReLU(),
                #Conv1D(128,3,padding='same'),
                #LeakyReLU(),
                MaxPool1D(2),
                Flatten(),
                Dense(512),
            ])
        )



from .asn_info import ASNInfo
from .clumps_flow import ClumpsFlow
from .clumps_flow_timeframe import ClumpsFlowTimeFrame
from .dns_counter import DNSCounter
from .first_packet_payload import FirstPacketPayloadLen
from .flowpic import FlowPic
from .graypic import GrayPic1
from .most_freq_payload_len_ratio import MostFreqPayloadLenRatio
from .n_bytes_per_packet import NBytesPerPacket
from .n_bytes import NBytes
from .n_pkts_byte_freq import NPacketsByteFrequency
from .packets_size_interarrival_time import PacketsSizeAndIAT
from .pkt_rel_time import PacketRelativeTime
from .protocol_header_fields import ProtocolHeaderFields
from .protocol_header_fields_extended import ProtocolHeaderFieldsExtended
from .protocol_header_fields_TDL import ProtocolHeaderFieldsTDL
from .recv_sent_pkt_ratio import RecvSentPacketRatio
from .res_req_diff_time import ResReqDiffTime
from .simple_tig import SimpleTIG
from .small_pkt_payload_ratio import SmallPacketPayloadRatio
from .stnn_extended import STNNExtended
from .stnn_timeframe import STNNTimeFrame
from .stnn import STNN
from .tdl import TDL
from .tgnn import TGNN

__all__ = [ # short for the plugins
    'ASNInfo',
    'ClumpsFlow',
    'ClumpsFlowTimeFrame',
    'DNSCounter',
    'FirstPacketPayloadLen',
    'FlowPic',
    'GrayPic1',
    'MostFreqPayloadLenRatio',
    'NBytesPerPacket',
    'NBytes',
    'NPacketsByteFrequency',
    'PacketsSizeAndIAT',
    'PacketRelativeTime',
    'ProtocolHeaderFields',
    'ProtocolHeaderFieldsExtended',
    'ProtocolHeaderFieldsTDL',
    'RecvSentPacketRatio',
    'ResReqDiffTime',
    'SimpleTIG',
    'SmallPacketPayloadRatio',
    'STNNExtended',
    'STNNTimeFrame',
    'STNN',
    'TDL',
    'TGNN'
]



from nfstream import NFPlugin
import numpy as np
from runstats import *
from scapy.all import IP, IPv6, raw

class TGNN(NFPlugin):
    ''' TGNN | A simple graph feature extractor.

    From each packet the plugin extracts the packet bytes, packet inter-arrival time, packet payload length, packet length. Also extract the flow duration.
    The packets bytes are used as weight for the nodes of the graph, the packet inter-arrival time is used as weight for the edges of the graph,
    and payload length mean, payload length SD, packet length mean, packet length SD, and flow duration are the meta features of the graph.
    
    Why simple: The plugin doesnt support any packet/node feature customization.
    
    Output Features: 
        1. `udps.simple_tgnn_adj`: Adjacency matrix of the graph. Shape of (`n_packets`, `n_packets`). 
        2. `udps.simple_tgnn_vertcies_features`: Node features, the packet size. Shape of (`n_packets`, 1).
        3. `udps.simple_tgnn_edges_weights`: Node weight edges, the packet inter-arrival time. Shape of (edges, 1).
        4. `udps.simple_tgnn_meta_features`: Graph meta features, payload length mean, payload length SD, packet length mean, packet length SD, and flow duration. Shape of (5).
        
    Paper:
        "Flow-based Encrypted Network Traffic Classification with Graph Neural Networks."
        
        By:
            - Ting-Li Huoh, 
            - Yan Luo, 
            - Peilong Li, 
            - Tong Zhang.
    '''
    def __init__(self, n_packets=5, MTU=1500):
        '''
        Args:
            `n_packets` (int): The number of packets to process per flow.
                flows with less than the desired `n_packets` will be padded with 
                unconnected zero-valued nodes to match `n_packets`.
        '''
        self.n_packets = n_packets
        self.MTU = MTU
    
    def on_init(self, packet, flow):
        '''
        on_init
        '''
        flow.udps.simple_tgnn_nodes_weight = np.zeros((self.n_packets,self.MTU))
        flow.udps.simple_tgnn_timestamp = np.zeros((self.n_packets))
        flow.udps.simple_tgnn_payload_length = Statistics()
        flow.udps.simple_tgnn_packet_length = Statistics()

        flow.udps.simple_tgnn_nodes_weight_no_norm = np.zeros((self.n_packets,self.MTU))
        self.on_update(packet, flow)

    def on_update(self, packet, flow):
        '''on_update
        '''
        if flow.bidirectional_packets > self.n_packets:
            return
        
        if packet.protocol in [6, 17]:
            # nodes weights
            copied_binary_payload = []
            copied_binary_payload.extend(self.get_packet_as_binary_pypacker(packet))
            padding = self.MTU-len(copied_binary_payload)
            copied_binary_payload.extend(np.full(padding,0))
            flow.udps.simple_tgnn_nodes_weight_no_norm[flow.bidirectional_packets-1] = copied_binary_payload
            copied_binary_payload = [float(i)/255 for i in copied_binary_payload]
            # copied_binary_payload.extend(np.full(padding,0))
            flow.udps.simple_tgnn_nodes_weight[flow.bidirectional_packets-1] = copied_binary_payload

            # Nodes timestamps for Edge inter-arrival time weights
            packet_timestamp = packet.time
            flow.udps.simple_tgnn_timestamp[flow.bidirectional_packets-1] = packet_timestamp
            
            # Graph meta atttributes
            payload_length = packet.payload_size
            packet_length = packet.ip_size

            flow.udps.simple_tgnn_payload_length.push(payload_length)
            flow.udps.simple_tgnn_packet_length.push(packet_length)




        
        
    def on_expire(self, flow):
        '''on_expire
        '''
        flow.udps.simple_tgnn_graph_meta_attr = [flow.udps.simple_tgnn_payload_length.mean(),
                           flow.udps.simple_tgnn_payload_length.stddev() if flow.udps.simple_tgnn_payload_length._count >= 2 else 0,
                           flow.udps.simple_tgnn_packet_length.mean(),
                           flow.udps.simple_tgnn_packet_length.stddev() if flow.udps.simple_tgnn_packet_length._count >= 2 else 0,
                           flow.bidirectional_duration_ms/1000]
        flow.udps.simple_tgnn_graph_meta_attr_no_norm = flow.udps.simple_tgnn_graph_meta_attr
        flow.udps.simple_tgnn_graph_meta_attr = [float(i)/max(flow.udps.simple_tgnn_graph_meta_attr) for i in flow.udps.simple_tgnn_graph_meta_attr]
        flow.udps.simple_tgnn_timestamp[flow.udps.simple_tgnn_timestamp == 0] = flow.udps.simple_tgnn_timestamp.max()
        graph = generate_simple_GNN(self.n_packets,flow.udps.simple_tgnn_timestamp)
        flow.udps.simple_tgnn_edges_weight = graph.edges_weight.tolist()
        flow.udps.simple_tgnn_sender_vector = graph.sender_vector.tolist()
        flow.udps.simple_tgnn_receiver_vector = graph.receiver_vector.tolist()
        flow.udps.simple_tgnn_timestamp = flow.udps.simple_tgnn_timestamp.tolist()
        flow.udps.simple_tgnn_nodes_weight = flow.udps.simple_tgnn_nodes_weight.tolist()
        del flow.udps.simple_tgnn_payload_length
        del flow.udps.simple_tgnn_packet_length
    
    def get_packet_as_binary_pypacker(self, packet):
        '''
        Scapy has a bug with correctly dissecting SNMP payload.
        '''
        return raw(packet.ip_packet)
    
    def get_payload_as_binary(self, packet):
        return packet.ip_packet[-packet.payload_size:]
    
    def get_payload_as_binary_scapy(self, packet):
        '''
        Older versions of NFStream (lower than 6.1) has problems with 
        detection the correct payload size of ipv6 packets.
        This function is here as a fallback in case of problems in future
        versions.

        in 6.4.2 and 6.4.3 there is a bug with wrong transport_size, also
        NFStream does not detect ethernet padding.
        '''
        ip_version = packet.ip_version
        if ip_version == 4:
            scapy_packet = IP(packet.ip_packet)
        elif ip_version == 6:
            scapy_packet = IPv6(packet.ip_packet)

        return raw(scapy_packet.payload.payload)
    
    @staticmethod
    def preprocess(dataframe):
        ''' 
        Preprocessing method for the GraphDApp features.
        Converting 'udps.simple_tig_adj' and 'udps.simple_tig_features' columns from str to 2D-list.
        '''
        import ast
        # validate
        assert 'udps.simple_tgnn_nodes_weight' in dataframe.columns, "Column 'udps.simple_tgnn_nodes_weight' not found."
        assert isinstance(dataframe['udps.simple_tgnn_nodes_weight'].iloc[0], str), "Values in column 'udps.simple_tgnn_nodes_weight' are already processed."

        dataframe['udps.simple_tgnn_nodes_weight'] = dataframe['udps.simple_tgnn_nodes_weight'].apply(ast.literal_eval)

        assert 'udps.simple_tgnn_edges_weight' in dataframe.columns, "Column 'udps.simple_tgnn_edges_weight' not found."
        assert isinstance(dataframe['udps.simple_tgnn_edges_weight'].iloc[0], str), "Values in column 'udps.simple_tgnn_edges_weight' are already processed."
        
        dataframe['udps.simple_tgnn_edges_weight'] = dataframe['udps.simple_tgnn_edges_weight'].apply(ast.literal_eval)



def generate_simple_GNN(nodes,nodes_connection_values):
    graph = Graph(nodes, nodes_connection_values)
    graph.chronologic_relationship()
    return graph


class Graph:
    def __init__(self, n, timestamps):
        self.n = n
        self.timestamps = timestamps
        self.edge_num = int((self.n*(self.n-1))/2)
        self.sender_vector = np.zeros((self.edge_num))
        self.receiver_vector = np.zeros((self.edge_num))
        self.edges_weight = np.zeros((self.edge_num,1))
    
    def connect(self, v1, v2, edge_num):
        self.sender_vector[edge_num] = v1
        self.receiver_vector[edge_num] = v2
        self.edges_weight[edge_num] = self.timestamps[v2] - self.timestamps[v1]
    
    def chronologic_relationship(self):
        for i in range(0,self.n-1):
            self.edges_weight[i] = self.timestamps[i+1]-self.timestamps[i]
            self.sender_vector[i] = i
            self.receiver_vector[i] = i + 1

        count = self.n-1
        for i in range(0,self.n):
            for num,j in enumerate(range(i+2,self.n)):
                self.connect(i,j,count+num)
            count = count + num + 1



from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input,Dense,Dropout

class G2VModality:
    '''
    G2V deep learning nerual network uses for Graph input.
    Classifciation type: Graphs
    Input: Graph2Vec vector 1-dimension matching the input shape of the model

    Graph2Vec algorithm can be found in preprocess

    Inspired by the paper: " graph2vec: Learning Distributed Representations of Graphs "
    Authors: Narayanan, Annamalai and Chandramohan, Mahinthan and Venkatesan, Rajasekar and Chen, Lihui and Liu, Yang and Jaiswal, Shantanu
    '''
    def __init__(self,input_shape) -> None:
        self.model = _G2V_Model(input_shape)



class _G2V_Model(Model):
    def __init__(self,input_shape) -> None:
        super(_G2V_Model, self).__init__(name="Graph2Vec_modality")

        #Layers
        self.input_layer = Input(shape=input_shape,name=f'input_vector_graph')
        self.dense = Dense(128,activation='relu')
        self.dropout = Dropout(0.2)
        self.dense_2 = Dense(128,activation='relu')
        self.dropout_2 = Dropout(0.2)
        self(inputs=self.input_layer)


    def call(self,inputs):
        dense_out =  self.dense(inputs)
        dropout_out = self.dropout(dense_out)
        dense_out_2 = self.dense_2(dropout_out)
        dropout_out_2 = self.dropout_2(dense_out_2)
        return dropout_out_2



from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input,GRU,ReLU,Bidirectional,Flatten
from tensorflow.keras.constraints import max_norm 
from ._utils import stack

class LopezExtendedModality:
    '''
    Lopez-Protocol-header-Fields Exteneded deep learning nerual network uses for Packet Header Fields vector input.
    Classifciation type: Packet Header Fields
    Input: two-dimensional vector of number of packets and 6 features from each packet.

    Extraction plugin: ProtocolHeaderFieldsExtended
    
    Inspired by the paper: "Network Traffic Classifier With Convolutional and Recurrent Neural Networks for Internet of Things"
    '''
    def __init__(self,packet_count=32) -> None:
        input_layer_protocol_fields_modality = Input(shape=(packet_count,6), name='input_protocol_fields')
        self.model = Model(
            name='Lopez protocol header fields extended modality',
            inputs=input_layer_protocol_fields_modality,
            outputs=stack([
                input_layer_protocol_fields_modality,
                Bidirectional(GRU(64, return_sequences=True, kernel_constraint=max_norm(3))),
                ReLU(),
                Flatten(),
            ])
        )



import librosa
import scipy
import numpy as np
from ._modal_wrapper import ModalWrapper
from sklearn.ensemble import RandomForestClassifier

class STFTModality:
    '''
    STFT is a deep learning nerual network which uses waves signals with short-time-fourier-transform (STFT) of the packets flow as input.
    Classifciation type: Time Series
    Input: runs on 15000 miliseconds using time and direction as it features

    Contributor: Natan Dilbary
    '''
    def __init__(self,input_shape) -> None:
        self.model = ModalWrapper(RandomForestClassifier(),name='short-time-fourier-transform_modality',input_shape=input_shape)

    def fit_transform(self,x):
        return self.transform_signals2STFT(x)
    
    def extract_STFT(self,data,features_to_extract=None):
        # Compute the STFT of the signal
        f, t, Zxx = scipy.signal.stft(data, fs=1000, window='hann', nperseg=256, noverlap=128)
        # Compute the magnitude spectrogram
        mag_spectrogram = np.abs(Zxx)

        # If no features are specified, use default features
        if features_to_extract is None:
            features_to_extract = ('mean', 'std', 'spectral_centroid',
                                'spectral_bandwidth', 'spectral_contrast',
                                'spectral_flatness','spectral_rolloff', 'chroma_stft', 'mfcc')

        feature_map = {
            'mean': lambda: np.mean(mag_spectrogram, axis=1),
            'std': lambda: np.std(mag_spectrogram, axis=1),
            'spectral_centroid': lambda: librosa.feature.spectral_centroid(S=mag_spectrogram)[0],
            'spectral_bandwidth': lambda: librosa.feature.spectral_bandwidth(S=mag_spectrogram)[0],
            'spectral_contrast': lambda: librosa.feature.spectral_contrast(S=mag_spectrogram)[0],
            'spectral_flatness': lambda: librosa.feature.spectral_flatness(S=mag_spectrogram)[0],
            'spectral_rolloff': lambda: librosa.feature.spectral_rolloff(S=mag_spectrogram, sr=1000)[0],
            'chroma_stft': lambda: librosa.feature.chroma_stft(S=mag_spectrogram, sr=1000)[0],
            'mfcc': lambda: librosa.feature.mfcc(S=mag_spectrogram, sr=1000)[0]
        }
        
        # Compute the requested features
        computed_features = [feature_map[feat]() for feat in features_to_extract if feat in feature_map]

        # Combine all computed features into one vector
        feature_vec = np.concatenate(computed_features)
        return feature_vec
    
    def transform_signals2STFT(self,signals):
        stft_list = []
        for signal in signals:
            features = np.array([self.extract_STFT(signal)])
            stft_list.append(features)
        return np.concatenate(stft_list)
    


from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Layer, BatchNormalization, MaxPool1D, Conv1D, ReLU, MaxPooling1D, Flatten, Bidirectional, GRU, LeakyReLU, Lambda, Dense, MaxPool2D, Conv2D, LSTM, Concatenate, Dropout
from tensorflow.keras.constraints import max_norm 
from tensorflow import expand_dims
from tensorflow.keras import activations
import tensorflow as tf
from . import CustomDistiller




class Hybrid(CustomDistiller):
    def __init__(self, payload_size=784, header_fields_packets=32, n_classes=[],merging_method='feat_concat') -> None:
        super(Hybrid, self).__init__(  
            modalities=[
                wang_payload_modality(payload_size),
                lopez_protocol_header_fields_extended_modality(header_fields_packets),
                stnn_inspired_image_extended_modality(),
                graphdapp_modality()
            ],
            adapter_size=128, 
            n_classes=n_classes,
            merging_method=merging_method
        )



def wang_payload_modality(payload_size=784):
    input_layer_payload_modality = Input(shape=(payload_size,1), name='input_payload')
    return Model(
        name='Wang payload modality - nbytes',
        inputs=input_layer_payload_modality,
        outputs=stack([
            input_layer_payload_modality,
            Conv1D(16, 25, name='Conv1D_payload_1'),
            ReLU(),
            MaxPooling1D(3, name='MaxPooling1D_payload_1'),
            Conv1D(32, 35, name='Conv1D_payload_2'),
            ReLU(),
            MaxPooling1D(3, name='MaxPooling1D_payload_2'),
            Flatten(), 
        ])
    )


def lopez_protocol_header_fields_modality(packet_count=32):
    input_layer_protocol_fields_modality = Input(shape=(packet_count,4), name='input_protocol_fields')
    return Model(
        name='Lopez protocol header fields modality',
        inputs=input_layer_protocol_fields_modality,
        outputs=stack([
            input_layer_protocol_fields_modality,
            Bidirectional(GRU(64, return_sequences=True, kernel_constraint=max_norm(3))),
            ReLU(),
            Flatten(),
        ])
    )

def lopez_protocol_header_fields_extended_modality(packet_count=32):
    input_layer_protocol_fields_modality = Input(shape=(packet_count,6), name='input_protocol_fields')
    return Model(
        name='Lopez protocol header fields extended modality',
        inputs=input_layer_protocol_fields_modality,
        outputs=stack([
            input_layer_protocol_fields_modality,
            Bidirectional(GRU(64, return_sequences=True, kernel_constraint=max_norm(3))),
            ReLU(),
            Flatten(),
        ])
    )

def stnn_inspired_image_modality():
    input_layer_stnn_modality = Input(shape=(5,14), name='input_stnn')
    return Model(
        name='STNN-inspired image modality',
        inputs=input_layer_stnn_modality,
        outputs=stack([
            input_layer_stnn_modality,
            Bidirectional(LSTM(65,return_sequences=True)),
            Lambda(lambda x: expand_dims(x, axis=3)),
            Conv2D(32,3,padding='same'),
            LeakyReLU(),
            Conv2D(32,3,padding='same'),
            LeakyReLU(),
            MaxPool2D(2),
            Conv2D(64,3,padding='same'),
            LeakyReLU(),
            Conv2D(128,3,padding='same'),
            LeakyReLU(),
            MaxPool2D(2),
            Flatten(),
            Dense(512),
        ])
    )

def stnn_inspired_image_extended_modality():
    input_size = 177-42
    input_layer_stnn_modality = Input(shape=(input_size,1), name='input_stnn')
    return Model(
        name='STNN-inspired image extended modality',
        inputs=input_layer_stnn_modality,
        outputs=stack([
            input_layer_stnn_modality,
            Bidirectional(LSTM(64,return_sequences=True)), #128
            Conv1D(32,3,padding='same'),
            LeakyReLU(),
            #Conv1D(32,3,padding='same'),
            #LeakyReLU(),
            #MaxPool1D(2),
            Conv1D(64,3,padding='same'),
            LeakyReLU(),
            #Conv1D(128,3,padding='same'),
            #LeakyReLU(),
            MaxPool1D(2),
            Flatten(),
            Dense(512),
        ])
    )


def graphdapp_modality():
    input_layer_graphdapp_adj_matrix_modality = Input(shape=(32,32),name='input_graphdapp_adj_matrix')
    input_layer_graphdapp_features_modality = Input(shape=(32,1),name='input_graphdapp_features')
    inputs_layer = [input_layer_graphdapp_adj_matrix_modality, input_layer_graphdapp_features_modality]
    mlp1 = MLPLayer(64)
    mlp2 = MLPLayer(64)
    mlp3 = MLPLayer(64)
    readout = Readout()
    concat = Concatenate()
    x1 = mlp1(inputs_layer)
    x2 = mlp2(x1)
    x3 = mlp3(x2)
    x4 = readout([x1, x2, x3])
    x4 = concat(x4)
    return Model(
        name='GraphDApp modality',
        inputs=inputs_layer,
        outputs= x4
    )

class Readout(Layer):
    def __init__(self):
        super(Readout, self).__init__()
    
    def call(self, inputs):
        FsBatch = []
        for i in range(len(inputs)):
            Fbatch = inputs[i][1]
            FsBatch.append(tf.reduce_sum(Fbatch, axis=-2))
        return FsBatch


class MLPLayer(Layer):
    def __init__(
            self, 
            output_size=5, 
            activation='relu', 
            use_bias=True, 
            neighbour_agg_method='sum', 
            dropout_rate=0.025):
        super(MLPLayer, self).__init__()
        self.output_size = output_size
        self.dense = Dense(self.output_size, use_bias=use_bias)
        self.activation = activations.get(activation)
        self.batch_norm = BatchNormalization()
        self.dropout = Dropout(dropout_rate)
    
    def build(self, input_shape):
        self.eps = self.add_weight(
            name='epsilon',
            shape=(1,),
            initializer="random_normal",
            dtype='float32',
            trainable=True,
        )
    
    def call(self, inputs, training=False):
        '''
        [A, F]
        '''
        A = inputs[0]
        F = inputs[1]
        outputs = tf.multiply(1.0+self.eps, F) + tf.matmul(A,F)
        outputs = self.dense(outputs)
        outputs = self.activation(outputs)
        outputs = self.batch_norm(outputs, training=training)
        outputs = self.dropout(outputs)
        return [A, outputs]


# TDL verions

def lopez_protocol_header_fields_TDL_modality(packet_count=32):
    input_layer_protocol_fields_modality = Input(shape=(packet_count,3), name='input_protocol_fields_TDL')
    return Model(
        name='Lopez protocol header fields modality',
        inputs=input_layer_protocol_fields_modality,
        outputs=stack([
            input_layer_protocol_fields_modality,
            Bidirectional(GRU(64, return_sequences=True, kernel_constraint=max_norm(3))),
            ReLU(),
            Flatten(),
        ])
    )

def stack(layers):
    '''
    Using the Functional-API of Tensorflow to build a sequential
    network (stacked layers) from list of layers.
    '''
    layer_stack = None
    for layer in layers:
        if layer_stack is None:
            layer_stack = layer
        else:
            layer_stack = layer(layer_stack)
    return layer_stack



# Modals
from ._graphdapp import GraphDAppModality
from ._modal_wrapper import ModalWrapper
from ._lopez_protocol_header_fields import LopezModality
from ._lopez_protocol_header_fields_extended import LopezExtendedModality
from ._lopez_protocol_header_fields_TDL import LopezTDLModality
from ._stnn import STNNModality
from ._stnn_extended import STNNExtendedModality
from ._wang_payload import WangPayloadModality
from ._graph2vec import G2VModality
from ._stft import STFTModality
from ._wavelet import WaveletModality

# utils
from ._utils import create_signal

__all__ = [ # Modals
    'LopezModality',
    'LopezExtendedModality',
    'LopezTDLModality',
    'STNNModality',
    'STNNExtendedModality',
    'WangPayloadModality',
    'G2VModality',
    'STFTModality',
    'WaveletModality',
    'ModalWrapper'
] + [ # utils func
    'create_signal'
]




from __future__ import annotations
import networkx as nx
import numpy as np
import random
from math import copysign
from gensim.models.doc2vec import Doc2Vec, TaggedDocument



class Graph2Vec:
    def __init__(self, wl_iteration: int = 2, dimensions: int = 128, epochs: int = 20, workers: int = 4, learning_rate: float = 0.025, min_count: int = 5, down_sample: float = 0.0001, seed: int = 42, ignore_attributes: bool = False, use_node_attributes: bool = True, use_edge_attributes: bool = False):
        '''
        wl_iteration: the number of weisfiler-lehman iteration to use on each root node in the graph.
        dimensions: the size of the output dimension of the vector create from the graph.
        epochs: the number of epochs to train the skip-gram model.
        workers: the number of workers to use in the skip-gram model.
        learning_rate: the rate of the learning for the skip-gram model.
        min_count: the number to ignores all words with total frequncy lower than that.
        down_sample: the threshold for configuring which higher-frequency words are randomly downsampled.
        seed: the random seed to set for the random variables in the graph2vec model.
        ignore_attributes: Boolean True/False value for the usage of graph attributes as part of the method to build weisfiler-lehman graph.
        use_node_attributes: Boolean True/False value for the usage of node attributes as part of the method to build weisfiler-lehman graph.
        use_edge_attributes: Boolean True/False value for the usage of edge attriubtes as part of the method to build weisfiler-lehman graph.
        
        The algorithm is inspired by the paper:
            "Graph2vec: Learning Distributed Representations of Graphs"

        Authors:
            - Annamalai Narayanan
            - Mahinthan Chandramohan
            - Rajasekar Venkatesan
            - Lihui Chen
            - Yang Liu
            - Shantanu Jaiswal

        The code is base on the karateclub graph2vec:
            https://github.com/benedekrozemberczki/karateclub/blob/master/karateclub/graph_embedding/graph2vec.py

        NOTE: Customized Graph2Vec algorithm allowing the use of weights on nodes and edges of the graph
        '''
        self.wl_iteration = wl_iteration
        self.dimensions = dimensions
        self.epochs = epochs
        self.workers = workers
        self.learning_rate = learning_rate
        self.min_count = min_count
        self.down_sample = down_sample
        self.seed = seed
        self.ignore_attributes = ignore_attributes
        self.use_node_attributes = use_node_attributes
        self.use_edge_attributes = use_edge_attributes


    
    def fit(self, graphs: list[nx.classes.Graph]):
        '''
        graphs: list of networkx graphs

        The function given the set of graphs as input and learn to transform into list of vectors.
        Each vector reperesnt a graph which was transformed using weisfiler-lehman and skip-grams.
        '''
        self._set_seed()

        documents = [
            self.weisfeiler_lehman(graph,
                                   wl_iteration=self.wl_iteration,
                                   use_node_attributes=self.use_node_attributes,
                                   use_edge_attributes=self.use_edge_attributes,
                                   ignore_attributes=self.ignore_attributes) for graph in graphs
        ]
        documents = [TaggedDocument(words=doc.split('_'),tags=[str(i)]) for i,doc in enumerate(documents)]

        self.model = Doc2Vec(
                            documents,
                            vector_size=self.dimensions,
                            window=0,
                            min_count=self.min_count,
                            dm=0,
                            sample=self.down_sample,
                            workers=self.workers,
                            epochs=self.epochs,
                            alpha=self.learning_rate,
                            seed=self.seed
                            )

        self._embedding = [self.model.dv[str(i)] for i, _ in enumerate(documents)]

    def infer(self, graphs: list[nx.classes.Graph]):
        '''
        graphs: list of networkx graphs

        The function given the set of graphs as input, transform and return list of vectors.
        '''
        self._set_seed()
    
        documents = [
            self.weisfeiler_lehman(graph,
                                   wl_iteration=self.wl_iteration,
                                   use_node_attributes=self.use_node_attributes,
                                   use_edge_attributes=self.use_edge_attributes,
                                   ignore_attributes=self.ignore_attributes) for graph in graphs
        ]

        embedding = np.array(
            [
            self.model.infer_vector(
                doc.split('_'),alpha=self.learning_rate,min_alpha=0.00001,epochs=self.epochs
            ) for doc in documents
            ]
        )
        return embedding

    def weisfeiler_lehman(self, graph: nx.classes.Graph, wl_iteration: int, use_node_attributes: bool, use_edge_attributes: bool, ignore_attributes: bool):
        '''
        graph: networkx graph
        wl_iteration: the number of weisfiler-lehman iteration to use on each root node in the graph.
        use_node_attributes: Boolean True/False value for the usage of node attributes as part of the method to build weisfiler-lehman graph.
        use_edge_attributes: Boolean True/False value for the usage of edge attriubtes as part of the method to build weisfiler-lehman graph.
        ignore_attributes: Boolean True/False value for the usage of graph attributes as part of the method to build weisfiler-lehman graph.

        NOTE: Customized weisfeiler lehman algorithm allowing the use of weights on nodes and edges of the graph
        '''
        canonical_form = [0]*len(graph.nodes())
        
        if ignore_attributes:
            use_node_attributes = False
            use_edge_attributes = False

        if use_node_attributes:
            feature_vec = [graph.nodes[node]['weight'] for node in graph.nodes()]
        else:
            feature_vec = [1]*len(graph.nodes())

        feature_vec = np.array(feature_vec)

        for _ in range(wl_iteration):
            for node in graph.nodes():
                labels = []
                nebs = graph.neighbors(node)
                for neb in nebs:
                    value = feature_vec[neb]
                    if use_edge_attributes:
                        if isinstance(value,np.ndarray):
                            value = value + copysign(graph.edges[(node,neb)]['weight'],value[0])
                        else:
                            value = value + copysign(graph.edges[(node,neb)]['weight'],value)
                    labels.append(value)

                value = labels[0]
                for i in range(1,len(labels)):
                    value += labels[i]
                canonical_form[node] = value
            feature_vec = canonical_form

        if isinstance(canonical_form[0],np.ndarray):
            new_canonical = []
            for value in canonical_form:
                for v in value:
                    new_canonical.append(v)
            canonical_form = new_canonical
            
        t = list(sorted(map(int,canonical_form),reverse=True))
        st = str(t[0])

        for i in range(1,len(t)):
            st += '_'+str(t[i])

        return st

    def _set_seed(self):
        random.seed(self.seed)
        np.random.seed(self.seed)
